#version 430    // core profile, compute shader

#ifdef COMPUTE_SHADER

uniform mat4 mvpMatrix;
uniform mat4 mvpvInvMatrix;

struct ray 
{
  vec3 o;
  vec3 d;
};

struct triangle
{
  vec3 p;
  vec3 edge1;
  vec3 edge2;
  vec4 color;

};

struct material
{
  vec3 diffuse_color;
  vec3 specular_color;
  vec3 emission;
  vec2 k;
};

struct source
{
  vec3 p;
  vec3 edge1;
  vec3 edge2;
  vec4 emission;
};

#define EPSILON 0.001
#define RAY_EPSILON 0.0001

bool intersect( in struct triangle triangle, in const ray ray, in const float htmax, out float rt ) //, out float ru, out float rv ) 
{
  /* begin calculating determinant - also used to calculate U parameter */
  vec3 pvec= cross(ray.d, triangle.edge2);
    
  /* if determinant is near zero, ray lies in plane of triangle */
  const float det= dot(triangle.edge1, pvec);
  //~ if (det > -EPSILON && det < EPSILON)
  //~ return false;
    
  const float inv_det= 1.0 / det;
    
  /* calculate distance from vert0 to ray origin */
  const vec3 tvec= ray.o - triangle.p;
  /* calculate U parameter and test bounds */
  const float u= dot(tvec, pvec) * inv_det;
  /* prepare to test V parameter */
  const vec3 qvec= cross(tvec, triangle.edge1);
  /* calculate V parameter and test bounds */
  const float v= dot(ray.d, qvec) * inv_det;
    
  /* calculate t, ray intersects triangle */
  rt= dot(triangle.edge2, qvec) * inv_det;
  //~ ru= u;
  //~ rv= v;
    
  // ne renvoie vrai que si l'intersection est valide (comprise entre tmin et tmax du rayon)
  return (u >= 0.0 && v >= 0.0 && u + v < 1.0 && rt < htmax && rt > RAY_EPSILON);
  //~ return true;
}

vec3 point(in struct triangle t, in float u,in float v ){
  float w= 1.f - u - v;
  return t.p * w + t.edge1 * u + t.edge2 * v;
}

float rand(vec2 co){
  return fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453);
}

float area(in struct triangle t){
  if(length(t.edge1) == 0.0 || length(t.edge2) == 0.0)
    return 0.0;
  return 0.5 * length(cross(t.edge1,t.edge2));
}


layout( binding = 0 ) readonly buffer triangleData  
{
  triangle triangles[];
};

layout( binding = 1 ) readonly buffer materialData
{
  material materials[];
};

layout( binding = 2 ) readonly buffer sourceData
{
  triangle sources[];
};

layout( binding= 0, rgba8 ) writeonly uniform image2D framebuffer;

layout( local_size_x= 32 , local_size_y= 2 ) in;
void main( )
{
  // rayon associe au pixel (local.x, local.y) de la tuile
  vec4 oh= mvpvInvMatrix * vec4(gl_GlobalInvocationID.xy, -1, 1);
  vec4 eh= mvpvInvMatrix * vec4(gl_GlobalInvocationID.xy, 1, 1);
    
  struct ray r;
  r.o= oh.xyz / oh.w;
  r.d= eh.xyz / eh.w - oh.xyz / oh.w;

  float h= 1.0;
  vec3 color;
  int i, j, k;
  // intersections avec les triangles
  float t;
  for(i= 0; i < triangles.length(); i++){
    if(intersect(triangles[i], r, h, t)){
      h= t;
      color= vec3(triangles[i].color);
      
      vec3 direct, indirect;
      //calcul du point d'intersection : o + d * t
      vec3 point = r.o + r.d * t;
      vec3 normal = normalize(cross(triangles[i].edge1,triangles[i].edge2));
      struct material m = materials[int(triangles[i].color.w)];


      //pour chaque source de lumière calculer l'énergie émise
      vec3 q;
      float pdf, u1, u2, s, t2, u, v;
      int n;
      for(j=0; j<sources.length(); j++ ){
	vec3 color_tmp;
	for(n=0; n<1; n++){

	  //calcul du point aléatoire sur la source
	  u1 = rand(vec2(0.0,1.0));
	  u2 = rand(vec2(0.0,1.0));
	  s = sqrt(u1);
	  t2 = u2;
	  u = 1.0 - s;
	  v = (1.0 - t2) * s;
	  pdf = 1.f / area(sources[j]);

	  float w= 1.f - u - v;
	  //	  q=sources[j].p * w + sources[j].edge1 * u + sources[j].edge2 * v;
	  q = sources[j].p;
	  //	  q = point(sources[j], u, v);

	  //tracer le rayon entre le point de la source et le point
	  struct ray r2;
	  r2.o = q;
	  r2.d = point-q;

	  float h2 = 1.0;
	  for(k= 0; k < triangles.length(); k++){
	    if(!intersect(triangles[k], r2, h2, t)){
	      h2 = t;
	      //calcul du point d'intersection : o + d * t
	      vec3 pointTmp = r2.o + r2.d * t;
	      vec3 normalTmp = normalize(cross(sources[i].edge1,sources[i].edge2));
	      float cosThetaP = max(0,dot(normal, normalize(-r2.d) ));
	      float cosThetaQ = max(0,dot(normalize(r2.d), normalTmp));
	      
	      float fr = m.k.x ;
	      
	      vec3 Li = vec3(sources[j].color.xyz);
	      float G = cosThetaP * cosThetaQ / (r2.d.x * r2.d.x + r2.d.y * r2.d.y + r2.d.z * r2.d.z);
	      color_tmp += Li * fr * G * 1.0 / pdf;
	    }
	  }
	}
	color_tmp /= 1.0;
	direct += color_tmp;
      }
      //      direct /= sources.length();



      color = normalize(direct);


      

      // color += indirect;
      

      //            color= triangles[i].color;
    }
  }
    
  color = normalize(color);
  imageStore(framebuffer, ivec2(gl_GlobalInvocationID.xy), vec4(color, 1));
}

#endif
