/**
 *
 * Copyright (C) 2013, 2014 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

#include <iostream>
#include <string>
#include <fstream>
#include <vector>
#include <utility>
#include <array>
#include <algorithm>
#include "App.h"
#include "Widgets/nvSDLContext.h"

#include "Mesh.h"
#include "MeshIO.h"

#include "Image.h"
#include "ImageIO.h"
#include "ImageArray.h"
#include "ImageManager.h"
#include "GLResource.h"

#include "GL/GLQuery.h"
#include "GL/GLTexture.h"
#include "GL/GLSampler.h"
#include "GL/GLFramebuffer.h"
#include "GL/GLBuffer.h"
#include "GL/GLVertexArray.h"
#include "GL/GLBasicMesh.h"
#include "ProgramManager.h"

#include "Grid.h"
#include "Camera.h"
#include "Plane.h"


#define SCENE 1 //numéro de la scène à charger (0|1)
#define DISTANCE_MAX_MAP 100 //distance maximum de chargement des données des blocks par rapport à la caméra
#define DISTANCE_MAX_PRINT 20 //distance maximum d'affichage des blocks par rapport à la caméra
#define LOADFULL 0 //initaliser le dos des faces

#define SCENE1_FILENAME "westeros/winterfell"
#define SCENE0_FILENAME "littlemap/littlemap"


#define GL_ARB_texture_cube_map_array 1

//////////////////////////////////////////////////////////////////
// UTILITAIRES
//////////////////////////////////////////////////////////////////

bool compareY (gk::Vector i,gk::Vector j) {
  return (i.y<j.y); 
}
bool compareX (gk::Vector i,gk::Vector j) {
  return (i.x<j.x); 
}

bool compareVisible (std::pair<int,double> i,std::pair<int,double> j) {
  return (i.second<j.second); 
}

int irand(int rangeLow, int rangeHigh) {
  return rangeLow+rand()%(rangeHigh-rangeLow);
}

double frand(double rangeLow, double rangeHigh) {
  return rangeLow+rand()*((rangeHigh-rangeLow)/(double)RAND_MAX);
}

//! classe utilitaire : permet de construire une chaine de caracteres formatee. cf sprintf.
struct Format
{
  char text[1024];
    
  Format( const char *_format, ... )
  {
    text[0]= 0;     // chaine vide
        
    // recupere la liste d'arguments supplementaires
    va_list args;
    va_start(args, _format);
    vsnprintf(text, sizeof(text), _format, args);
    va_end(args);
  }
    
  ~Format( ) {}
    
  // conversion implicite de l'objet en chaine de caracteres stantard
  operator const char *( )
  {
    return text;
  }
};


const char *block_names[] = 
  {
    "", // 0
    "stone",
    "grass",
    "dirt",
    "cobblestone",
    "plank",
    "sapling",
    "bedrock",
    "water_flow",
    "water_still",
    "lava_flow",
    "lava_still",
    "sand",
    "gravel",
    "gold_ore",
    "iron_ore",
    "coal_ore",
    "wood",     //
    "leaves",
    "sponge",
    "glass",
    "lapis_ore",
    "lapis_block",
    "dispenser",
    "sandstone",
    "note_block",
    "bed_block",
    "rail",
    "detector",
    "piston",
    "web",
    "deadbush",
    "grass",
    "fern",
    "deadbush",
    "piston",
    "piston_top",
    "wool_colored_white",
    "wool_colored_orange",
    "wool_colored_magenta",
    "wool_colored_light_blue",
    "wool_colored_yellow",
    "wool_colored_lime",
    "wool_colored_pink",
    "wool_colored_gray",
    "wool_colored_gray",        // light gray
    "wool_colored_cyan",
    "wool_colored_purple",
    "wool_colored_blue",
    "wool_colored_brown",
    "wool_colored_green",
    "wool_colored_red",
    "wool_colored_black",
    "flower_dandelion",
    "flower_rose",
    "mushroom_brown",
    "mushroom_red",
    "gold_block",
    "iron_block",
    "stone_slab",
    "sandstone",       // standstone_slab
    "wood", // wood_slab
    "cobblestone",      // cobblestone_slab
    "brick",    // brick slab
    "stonebrick",       // stonebrick slab
    "nether_brick",     // netherbrick_slab
    "quartz_block",     // quartz_slab
    "brick",
    "tnt",
    "bookshelf",
    "cobblestone_mossy",
    "obsidian",
    "torch_on",
    "fire_layer0",
    "mob_spawner",
    "wood", // wood stairs
    "wood",     // chest
    "redstone", // redstone_wire
    "diamond_ore",
    "diamond_block",
    "wood",     // workbench
    "wheat_stage_5",    // wheat crops
    "dirt",     // soil
    "furnace_off",
    "furnace_on",
    "wood", // sign post
    "door_wood",
    "ladder",
    "rail_normal",
    "cobblestone", // cobblestone stair
    "stone", // wall sign
    "lever",
    "stone", // sotne pressure plate
    "door_iron",
    "wood",     // wood pressure plate
    "redstone_ore",
    "redstone", // glowing redstone 
    "redstone_torch_off",
    "redstone_torch_on",
    "stone", // stone button
    "snow", 
    "ice",
    "snow", // snow_block
    "cactus",
    "clay",
    "grass", // sugar cane
    "jukebox",
    "fence",    // ??
    "pumpkin",
    "netherrack",
    "sould_sand",
    "glowstone",
    "portal",
    "pumpkin", // jack o lantern
    "cake",
    "redstone_torch_off",       // redstone repeater off
    "redstone_torch_on",       // redstone repeater on
    "wood",     // locked chest
    "trapdoor",
    "stone",    // silverfish
    "cobblestone",      // silver
    "stonebrick",       // silver
    "stonebrick",
    "stonebrick_mossy",
    "stonebrick_cracked",
    "mushroom_block_red",       // red mushroom cap
    "mushroom_block_brown",       // brown mushroom cap
    "iron_bars",
    "glass_pane",
    "melon",
    "pumpkin_stem_disconnected",
    "melon",    // melon stem
    "vine",
    "wood", // fence gate
    "brick",    // brick stairs
    "stonebrick",       // stone brick stairs
    "mycelium",
    "waterlily",
    "nether_brick",
    "nether_brick",     // fence
    "nether_brick",     // stairs
    "nether_wart_stage_2",
    "enchanting_table",
    "brewing_stand",
    "cauldron",
    "end_frame",   // end portal
    "end_stone", 
    "dragon_egg",
    "redstone_lamp_off",
    "redstone_lamp_on",
    "planks_oak",      // double oak wood slab
    "planks_spruce",
    "planks_birch",
    "planks_jungle",
    "planks_oak",      // oak wood slab
    "planks_spruce",
    "planks_birch",
    "planks_jungle",
    "cocoa_stage_1",
    
    NULL        // 128
    
  };

/**
 * \brief structure pour les sommets
 */
struct sommet{
  gk::Vec3 position;
  gk::Vec3 normal;
  gk::Vec4 texcoord;
};


/* float fov = 50.f;
   float aspect = 1.f;
   float near = 1.f;
   float far = 1000.f;*/
class TP : public gk::App
{
  nv::SdlContext m_widgets;
    
  gk::GLProgram *m_program;
  gk::GLProgram *m_shadow;
  gk::GLProgram *m_sky;
  GLuint m_vao;
  GLuint bsommets;
  GLuint bindex;
  GLuint depthFramebuffer;
  GLuint depthTexture;
  GLuint sky_cube_vertex;
  GLuint m_sky_vao;
  // GLuint textureMapArray;
  std::array<int,256> stats; 
  std::vector< std::pair<GLint,GLsizei> > v_index_size;   //start, size
  std::vector< std::array<gk::Vector,8> > v_pbox; //liste des points des boites englobantes
  std::vector< std::pair<int,double> > v_visibleBox; //liste des block visible à jour

  gk::GLTexture *skyTexture;
  gk::GLTexture *cubeTexture;
  gk::GLTexture *textures;

  gk::Point currentBlock;
  gk::BBox shadow_zone;

  Camera m_camera;

  gk::GLQuery *m_time;  
  int update; //indiquer s'il est necessaire de recalculer la visibilité
  int mousePress;
  int frame=0,time,timebase=0;
  double fps=0, distanceMaxMap, distanceMaxPrint;
  int full;
  
  World world;
  std::string filename;
  gk::TVec3<int> scale;
  gk::Vec3 light_position,  light_direction;
public:
  // creation du contexte openGL et d'une fenetre
  TP( )
    :
    gk::App(),
    m_program(NULL),
    m_shadow(NULL),
#if SCENE == 1
    m_camera(50.f, 1.f, 1.f, 1000.f, gk::Vector(-1120.f,100.f,-8160.f),gk::Vector(-800.f,-101.f,-8160.f),gk::Vector(0.f,1.f,0.f),0.6,0.2),
#else
    m_camera(50.f, 1.f, 1.f, 1000.f, gk::Vector(14.f,20.f,4.f),gk::Vector(0.f,-1.f,0.f),gk::Vector(0.f,1.f,0.f),1,0.2,200,-60),
#endif
    distanceMaxMap(DISTANCE_MAX_MAP),
    distanceMaxPrint(DISTANCE_MAX_PRINT),
    full(LOADFULL),
    light_position(10.5f, 25.9f, 0.8f),
    light_direction(-1.f, 1.f, 1.f){
#if SCENE == 1
    filename=SCENE1_FILENAME;
#else
    filename=SCENE0_FILENAME;
#endif

    // specifie le type de contexte openGL a creer :
    gk::AppSettings settings;
    settings.setGLVersion(4,3);     // version 3.3
    settings.setGLCoreProfile();      // core profile
    settings.setGLDebugContext();     // version debug pour obtenir les messages d'erreur en cas de probleme
    settings.setSwapControl(1);
    // cree le contexte et une fenetre
    if(createWindow(512, 512, settings) < 0)
      closeWindow();
        
    m_widgets.init();
    m_widgets.reshape(windowWidth(), windowHeight());
  }
    
  ~TP( ) {}



  gk::Point getPointBlock(const gk::Vector& v){

    int xW=(int)v.x/scale.x;
    int yW=(int)v.y/scale.y;
    int zW=(int)v.z/scale.z;
    
    int xM=(int)(v.x-xW)/(scale.x/16);
    int yM=(int)(v.y-yW)/(scale.y/1);
    int zM=(int)(v.z-zW)/(scale.z/16);

    int xR=(int)(v.x-xW-xM)/(scale.x/16/16);
    int yR=(int)(v.y-yW-yM)/(scale.y/1/16);
    int zR=(int)(v.z-zW-zM)/(scale.z/16/16);   
    gk::Point p(xR,yR,zR);

    return p;
  }

  int loadLocalMash(gk::Point position, double distanceMaxMap){
  
    glUseProgram(m_program->name);
    // charge un mesh    
    gk::Mesh *mesh= gk::MeshIO::readOBJ("cube.obj");
    if(mesh == NULL)
      return -1;
    gk::MeshIO::buildNormals(mesh);   
    /////////////////////////////////////////////////////
    
    // for(unsigned int i =0;i<mesh->texcoords.size();i++)
    //   std::cout<<i<<" ["<<mesh->texcoords[i].x<<";"<<mesh->texcoords[i].y<<";"<<mesh->texcoords[i].z<<"]"<<std::endl;

    //TEST : Réduction du maillage, le cube est fonctionnel, mais le maillage n'a pas été adapté.
    std::vector<gk::Vec3> positions(8);
    for(int i = 0;i<8;i++){			      
      positions[i] = i%2 ? gk::Vec3(0,0,0):gk::Vec3(1,1,1);
      if(i & 1u<<1)
	positions[i].x = i%2 ? 1: 0;
      if(i & 1u<<2)
	positions[i].y = i%2 ? 1: 0;
      if(i & 1u<<3)
	positions[i].z = i%2 ? 1: 0;				 
    }

    std::vector<gk::Vec3> normals(8);
    for(unsigned int i =0;i<positions.size();i++){
      gk::Vector(positions[i]).print();



    }

    stats.fill(0);
    sommet s1,s2,s3,s4;

    std::vector< sommet > sommets;
    std::vector< unsigned int > indices;
    std::array<gk::Vector,8> pbox;

    Gridpoint pW,pM,pR,pB,pTmp;
    int id_mesh = 0, nb_voxel=0,face, voxel,x,y,z,add;
    unsigned int index_first,index_count; 
    gk::BBox box;
   
    v_index_size.clear();
    v_pbox.clear();

    //    std::cout<<"A"<<std::endl;
    for(int xW= 0; xW < world.size.x; xW++) 
      for(int yW= 0; yW < world.size.y; yW++) 
	for(int zW= 0; zW < world.size.z; zW++){	  
	  pW = world.grid_point(xW, yW, zW) ; 
	  Map *map= world.map(pW);
	  //	  std::cout<<"B"<<std::endl;
	  if(map == NULL)
	    continue;
	  {
	    //parcours de la map	  
	    for(int xM= 0; xM < map->size.x; xM++) 
	      for(int yM= 0; yM < map->size.y; yM++) 
		for(int zM= 0; zM < map->size.z; zM++){
		  pM =  map->grid_point(xM, yM, zM) ;
		  Region *region= map->region(pM);
		  if(region == NULL)
		    continue;
		  {
		    //		    std::cout<<"C"<<std::endl;
		    //parcours de la région	  
		    for(int xR= 0; xR < region->size.x; xR++) 
		      for(int yR= 0; yR < region->size.y; yR++) 
			for(int zR= 0; zR < region->size.z; zR++){
			  pR =  region->grid_point(xR, yR, zR) ;
			  Block *block= region->block(pR);
			  if(block == NULL)
			    continue;		
			  {

			    gk::Vector pMin(
					    world.scale.x * xW + map->scale.x * xM + region->scale.x * xR + world.bbox.pMin.x,
					    world.scale.y * yW + map->scale.y * yM + region->scale.y * yR + world.bbox.pMin.y,
					    world.scale.z * zW + map->scale.z * zM + region->scale.z * zR + world.bbox.pMin.z);
			    gk::Vector pMax(
					    world.scale.x * xW + map->scale.x * xM + region->scale.x * (xR+1) + world.bbox.pMin.x,
					    world.scale.y * yW + map->scale.y * yM + region->scale.y * (yR+1) + world.bbox.pMin.y,
					    world.scale.z * zW + map->scale.z * zM + region->scale.z * (zR+1) + world.bbox.pMin.z);
			    			  
			    if(distanceMaxMap &&
			       (Distance(currentBlock,getPointBlock(pMin))>=distanceMaxMap ||
				Distance(currentBlock,getPointBlock(pMax))>=distanceMaxMap)){
			      continue;
			    }
			    // if(!indices.size())
			    {

			      index_count=0;
			      //parcours du block
			      for(int xB= 0; xB < block->size.x; xB++)
				for(int yB= 0; yB < block->size.y; yB++) 
				  for(int zB= 0; zB < block->size.z; zB++){
				    pB =  block->grid_point(xB, yB, zB) ;
				    voxel= block->voxel(pB);				  
				    if(voxel>0){	
				      stats[voxel]++;
				      x = world.scale.x * xW + map->scale.x * xM + region->scale.x * xR + zB + world.bbox.pMin.x;
				      y = world.scale.y * yW + map->scale.y * yM + region->scale.y * yR + yB + world.bbox.pMin.y;
				      z = world.scale.z * zW + map->scale.z * zM + region->scale.z * zR + xB + world.bbox.pMin.z;
				      int x2 = world.scale.x * xW + map->scale.x * xM + region->scale.x * xR + xB + world.bbox.pMin.x;
				      int y2 = world.scale.y * yW + map->scale.y * yM + region->scale.y * yR + yB + world.bbox.pMin.y;
				      int z2 = world.scale.z * zW + map->scale.z * zM + region->scale.z * zR + zB + world.bbox.pMin.z;
				     	
				      if( 
					 (xB>0 && xB<15 &&  block->voxel(block->grid_point(xB+1, yB, zB)) && block->voxel(block->grid_point(xB-1, yB, zB)) ) &&
					 (yB>0 && yB<15 &&  block->voxel(block->grid_point(xB, yB+1, zB)) && block->voxel(block->grid_point(xB, yB-1, zB)) ) &&
					 (zB>0 && zB<15 &&  block->voxel(block->grid_point(xB, yB, zB+1)) && block->voxel(block->grid_point(xB, yB, zB-1)) )
					  )
					continue;
				     

				      std::vector<gk::Vec3> pos(mesh->positions);
				      for(unsigned int i = 0; i<mesh->positions.size(); i++){
					pos[i].x += x;
					pos[i].y += y;
					pos[i].z += z;
				      }
				      face=0;

				      do{
					int xBTmp=xB,yBTmp=yB,zBTmp=zB,bord=0;
					face++;
					if(face>5) face=0;
				      
					if(face>0 and face<3){ // 1 && 2
					  zBTmp+=face%2?-1:1;
					}else if(face>2 and face<5){//3 &&  4
					  xBTmp+=face%2?-1:1;
					}else{ // 5 && 0
					  yBTmp+=face%2?-1:1;
					}
					add = face*4;

					bord = !xB || ! yB || !zB || xB == 15 || yB == 15 || zB == 15 ?1:0;
					if(!block->voxel(block->grid_point(xBTmp, yBTmp, zBTmp)) or bord){
					
					  if(bord){

					   

					    //si c'est le bord
					    if(face==5){//si c'est la face du dessous
					      if(yB==0){//si on est dans le voxel du bas
						//ne pas afficher la face si le voxel du haut en xz du block de dessous  est >0
						Block *bottom  = world.block(Gridpoint(x, y-16, z));
						if( bottom!=NULL){
						  pTmp = bottom->grid_point(xB, 15, zB);
						  if( bottom->voxel(pTmp)>0)
						    continue;
						}
					      }else{//sinon
						pTmp = block->grid_point(xB, yB-1, zB);
						if( block->voxel(pTmp)>0)
						  continue;
					      }
					    }
					    if(face==0){
					      if(yB==15){
						Block *top     = world.block(Gridpoint(x, y+16, z));
						if( top!=NULL){					
						  pTmp = top->grid_point(xB, 0, zB);
						  if( top->voxel(pTmp)>0 )
						    continue;
						}
					      }else{
						pTmp = block->grid_point(xB, yB+1, zB);
						if( block->voxel(pTmp)>0)
						  continue;
					      }
					    }
					    			    
					    if(face==3 ){//si c'est la face de gauche
					      if(xB==0){//si on est dans le voxel de gauche
						//ne pas afficher la face si le voxel de droite en yz du block de gauche est >0
						Block *left    = world.block(Gridpoint(x2, y2, z2-16));
						if( left!=NULL){
						  pTmp = left->grid_point(15, yB, zB);
						  if( left->voxel(pTmp)>0)
						    continue;
						}
					      }else{						
						pTmp = block->grid_point(xB-1, yB, zB);
						if( block->voxel(pTmp)>0)
						  continue;						
					      }
					    }
					    if(face==4){
					      if(xB==15){		
						Block *right   = world.block(Gridpoint(x2, y2, z2+16));
						if( right!=NULL){
						  pTmp = right->grid_point(0, yB, zB);
						  if(right->voxel(pTmp)>0)
						    continue;
						}
					      }else{						
						pTmp = block->grid_point(xB+1, yB, zB);
						if( block->voxel(pTmp)>0)
						  continue;						
					      }
					    }
					    
					    
					    if(face==1){
					      if(zB==0){
						Block *front   = world.block(Gridpoint(x2-16, y2, z2));
						if( front!=NULL){
						  pTmp = front->grid_point(xB, yB, 15);			     
						  if( front->voxel(pTmp)>0)
						    continue;
						}
					      }else{
						pTmp = block->grid_point(xB, yB, zB-1);
						if( block->voxel(pTmp)>0)
						  continue;
					      }
					    }
					    if(face==2){
					      if(zB==15){
						Block *back    = world.block(Gridpoint(x2+16, y2, z2));
						if( back!=NULL){
						  pTmp = back->grid_point(xB, yB, 0);
						  if( back->voxel(pTmp)>0 )
						    continue;
						}
					      }else{
						pTmp = block->grid_point(xB, yB, zB+1);
						if( block->voxel(pTmp)>0)
						  continue;
					      }
					    }
					  }	
					  int id = (voxel)*6+face;

					  s1.position=pos[add+0];
					  s2.position=pos[add+1];
					  s3.position=pos[add+2];
					  s4.position=pos[add+3];
					  
					  s1.normal=mesh->normals[add+0];
					  s2.normal=mesh->normals[add+1];
					  s3.normal=mesh->normals[add+2];
					  s4.normal=mesh->normals[add+3];

					  /* s1.texcoord=gk::Vec4(mesh->positions[add+0].x,mesh->positions[add+0].y,mesh->positions[add+0].z,voxel);
					     s2.texcoord=gk::Vec4(mesh->positions[add+1].x,mesh->positions[add+1].y,mesh->positions[add+1].z,voxel);
					     s3.texcoord=gk::Vec4(mesh->positions[add+2].x,mesh->positions[add+2].y,mesh->positions[add+2].z,voxel);
					     s4.texcoord=gk::Vec4(mesh->positions[add+3].x,mesh->positions[add+3].y,mesh->positions[add+3].z,voxel);*/
					
					
					  s1.texcoord=gk::Vec4(mesh->texcoords[0].x,mesh->texcoords[0].y,id,0);
					  s2.texcoord=gk::Vec4(mesh->texcoords[1].x,mesh->texcoords[1].y,id,0);
					  s3.texcoord=gk::Vec4(mesh->texcoords[2].x,mesh->texcoords[2].y,id,0);
					  s4.texcoord=gk::Vec4(mesh->texcoords[3].x,mesh->texcoords[3].y,id,0);


					  sommets.push_back(s1);
					  sommets.push_back(s2);
					  sommets.push_back(s3);
					  sommets.push_back(s4);

					  indices.push_back(sommets.size()-4);
					  indices.push_back(sommets.size()-3);
					  indices.push_back(sommets.size()-2);
					
					  indices.push_back(sommets.size()-4);
					  indices.push_back(sommets.size()-2);
					  indices.push_back(sommets.size()-1);
					    
					  if(full){
					    indices.push_back(sommets.size()-4);
					    indices.push_back(sommets.size()-2);
					    indices.push_back(sommets.size()-3);
					
					    indices.push_back(sommets.size()-4);
					    indices.push_back(sommets.size()-1);
					    indices.push_back(sommets.size()-2);

					    index_count +=6;
					  }

					  index_count +=6;
					}
					

				      }while(face>0);
				    
				   
				      nb_voxel++;
				    }//end voxel
				
				  }//end block	
			   
			      if(indices.size()){

			  
				index_first=indices.size()-index_count;
				v_index_size.push_back(std::make_pair((GLint)index_first,(GLsizei)index_count));			    
			

				//			      int xBlock=	      
				for(int i = 0;i<8;i++){			      
				  pbox[i] = i%2 ? pMin:pMax;
				  if(i & 1u<<1)
				    pbox[i].x = i%2 ? pMax.x: pMin.x;
				  if(i & 1u<<2)
				    pbox[i].y = i%2 ? pMax.y: pMin.y;
				  if(i & 1u<<3)
				    pbox[i].z = i%2 ? pMax.z: pMin.z;				 
				}
				v_pbox.push_back(pbox);
				id_mesh++;
			      }
			    }
			  }
			 
			}//end region
		  } 
		}//end map
	  }    
	}//end world	   
     
    glGenVertexArrays(1, &m_vao);
    glBindVertexArray(m_vao); 
    
    glGenBuffers(1, &bsommets);  
    glBindBuffer(GL_ARRAY_BUFFER, bsommets);

    glBufferData(GL_ARRAY_BUFFER, sommets.size() * sizeof(sommet), &sommets[0], GL_STATIC_DRAW);
        
    glEnableVertexAttribArray(0);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(sommet),  reinterpret_cast<void*>( (int)offsetof(struct sommet, position)) );
    glEnableVertexAttribArray(1);
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, sizeof(sommet),  reinterpret_cast<void*>( (int)offsetof(struct sommet, normal)));
    glEnableVertexAttribArray(2);
    glVertexAttribPointer(2, 4, GL_FLOAT, GL_FALSE, sizeof(sommet),  reinterpret_cast<void*>( (int)offsetof(struct sommet, texcoord)));
    
    glGenBuffers(1, &bindex);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bindex);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(unsigned int), &indices[0], GL_STATIC_DRAW);

    std::cout<<"Loading world END : "<<sommets.size()<<" vertex "<< indices.size()<<" indices "<<nb_voxel<<" voxel(s)"<<std::endl;
    delete mesh;
    return 0;
  }
   
  int initMesh (){

    if(world.loadMap(filename) < 0)
      return -1;
    scale = world.scale;
    currentBlock = getPointBlock(m_camera.getPosition());
    
    loadLocalMash(currentBlock, distanceMaxMap);

    return 0;
  }

  int initTextures(){


    int width, height;
    
    glUseProgram(m_program->name);
    std::string path("images/blocks/");
    gk::ImageArray* cubemap = new gk::ImageArray();
    gk::Image *defaultImage= gk::readImage(path+"default.tif");

    for(unsigned int i=1;i<sizeof(block_names)/sizeof(block_names[0]) ;i++){
      if(block_names[i]!=NULL ){
	std::string filename(block_names[i]);
	gk::Image *image= gk::readImage(path+filename+".tif");
	gk::Image *image_side= gk::readImage(path+filename+"_side.tif");
	gk::Image *image_top= gk::readImage(path+filename+"_top.tif");
	gk::Image *image_bottom= gk::readImage(path+filename+"_bottom.tif");
	if(image == NULL && ( image_top==NULL || image_side==NULL || image_bottom==NULL))
	  image = defaultImage;
	
	if(image!=NULL){
	  width = image->width;
	  height = image->height;
	}

	cubemap->push_back(image_top==NULL?image:image_top);//0
	cubemap->push_back(image_side==NULL?image:image_side);//1
	cubemap->push_back(image_side==NULL?image:image_side);//2
	cubemap->push_back(image_side==NULL?image:image_side);//3
	cubemap->push_back(image_side==NULL?image:image_side);//4
	cubemap->push_back(image_bottom==NULL?image:image_bottom);//5

      }
   
    }



    if(!cubemap->validate(width,height)){

      textures= (new gk::GLTexture())->createTexture2DArray(gk::GLTexture::UNIT0, cubemap);
      std::cout<<"CREATE OK"<<std::endl;
      GLuint TextureID  = glGetUniformLocation(m_program->name, "textures");
      glUniform1i(TextureID, 1);

    }else std::cout<<"fail"<<std::endl;
    
    /*
      Test des CUBE ARRAY MAP, inspiration du livre OpenGL programming guide 4.3, mais non fonctionnelle
    */
    /*
    glGenTextures(1, &textureMapArray);
    if(textureMapArray != 0)
      std::cout<<"OK2"<<std::endl;
    //glActiveTexture(gk::GLTexture::UNIT0);
    glBindTexture(GL_TEXTURE_CUBE_MAP_ARRAY, textureMapArray);
    
    gk::TextureFormat format( GL_RGBA, GL_RGBA, GL_UNSIGNED_BYTE );

    unsigned int nbTextures = cubemap.images.size()/6;
    for(unsigned int cube= 0; cube < nbTextures; cube++){
      for(unsigned int face= 0; face < 6; face++){
        GLenum data_format= 0;
        switch(cubemap.images[cube*6+face]->channels){
	  case 1: data_format= GL_RED; break;
	  case 2: data_format= GL_RG; break;
	  case 3: data_format= GL_RGB; break;
	  case 4: data_format= GL_RGBA; break;
	  default: assert(0 && "invalid image format");
	  }
	
        GLenum data_type= GL_UNSIGNED_BYTE;
        switch(cubemap.images[cube*6+face]->type) {
	case gk::Image::UNSIGNED_BYTE: data_type= GL_UNSIGNED_BYTE; break;
	case gk::Image::FLOAT: data_type= GL_FLOAT; break;
	default: assert(0 && "invalid image format");
	}

        glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, 0, 
		     format.internal, cubemap.images[0]->width, cubemap.images[0]->height, 0,
		     data_format, data_type, cubemap.images[cube*6+face]->data);
	
      }
    
    }
    glGenerateMipmap(GL_TEXTURE_CUBE_MAP_ARRAY);   
    */  
    std::cout<<"TEXTURES OK"<<std::endl;
    return 0;
  }
  int initSkyMap(){

    glUseProgram(m_sky->name);
    glGenVertexArrays(1, &m_sky_vao);
    glBindVertexArray(m_sky_vao); 
    gk::Vec3 SkyBoxVertices[36] = 
      {
	gk::Vec3( 1.0f, -1.0f, -1.0f), 
	gk::Vec3( 1.0f, -1.0f,  1.0f), 
	gk::Vec3( 1.0f,  1.0f,  1.0f), 
	gk::Vec3( 1.0f,  1.0f,  1.0f), 
	gk::Vec3( 1.0f,  1.0f, -1.0f), 
	gk::Vec3( 1.0f, -1.0f, -1.0f),
	gk::Vec3(-1.0f, -1.0f,  1.0f),
	gk::Vec3(-1.0f, -1.0f, -1.0f),
	gk::Vec3(-1.0f,  1.0f, -1.0f),
	gk::Vec3(-1.0f,  1.0f, -1.0f),
	gk::Vec3(-1.0f,  1.0f,  1.0f),
	gk::Vec3(-1.0f, -1.0f,  1.0f),
	gk::Vec3(-1.0f,  1.0f, -1.0f),
	gk::Vec3( 1.0f,  1.0f, -1.0f),
	gk::Vec3( 1.0f,  1.0f,  1.0f),
	gk::Vec3( 1.0f,  1.0f,  1.0f),
	gk::Vec3(-1.0f,  1.0f,  1.0f),
	gk::Vec3(-1.0f,  1.0f, -1.0f),
	gk::Vec3(-1.0f, -1.0f,  1.0f),
	gk::Vec3( 1.0f, -1.0f,  1.0f),
	gk::Vec3( 1.0f, -1.0f, -1.0f),
	gk::Vec3( 1.0f, -1.0f, -1.0f),
	gk::Vec3(-1.0f, -1.0f, -1.0f),
	gk::Vec3(-1.0f, -1.0f,  1.0f),
	gk::Vec3( 1.0f, -1.0f,  1.0f),
	gk::Vec3(-1.0f, -1.0f,  1.0f),
	gk::Vec3(-1.0f,  1.0f,  1.0f),
	gk::Vec3(-1.0f,  1.0f,  1.0f),
	gk::Vec3( 1.0f,  1.0f,  1.0f),
	gk::Vec3( 1.0f, -1.0f,  1.0f), 
	gk::Vec3(-1.0f, -1.0f, -1.0f),
	gk::Vec3( 1.0f, -1.0f, -1.0f),
	gk::Vec3( 1.0f,  1.0f, -1.0f),
	gk::Vec3( 1.0f,  1.0f, -1.0f),
	gk::Vec3(-1.0f,  1.0f, -1.0f),
	gk::Vec3(-1.0f, -1.0f, -1.0f) 
      };
    glGenBuffers(1, &sky_cube_vertex); 
    glBindBuffer(GL_ARRAY_BUFFER, sky_cube_vertex);
    glBufferData(GL_ARRAY_BUFFER, 432, SkyBoxVertices, GL_STATIC_DRAW);
    glVertexAttribPointer(m_sky->attribute("vertex"), 3, GL_FLOAT, GL_FALSE, 0, (void*)0);
    glEnableVertexAttribArray(m_sky->attribute("vertex"));
    /*
    // charge un mesh    
    gk::Mesh *mesh= gk::MeshIO::readOBJ("cube.obj");
    if(mesh == NULL)
    return -1;
    gk::MeshIO::buildNormals(mesh);   
    
    m_sky_vertex_buffer= gk::createBuffer(GL_ARRAY_BUFFER, mesh->positions);
    glVertexAttribPointer(m_sky->attribute("vertex"), 3, GL_FLOAT, GL_FALSE, 0, 0);
    glEnableVertexAttribArray(m_sky->attribute("vertex"));
 
    m_sky_index_buffer= gk::createBuffer(GL_ELEMENT_ARRAY_BUFFER, mesh->indices);
    m_sky_indices_size= mesh->indices.size();
    */

    gk::ImageArray *envmap= gk::readImageArray("images/sky%02d.tif", 6);        
    skyTexture= (new gk::GLTexture())->createTextureCube(gk::GLTexture::UNIT0, envmap);

    GLuint TextureID  = glGetUniformLocation(m_sky->name, "cubemap");
    glUniform1i(TextureID, 5);
    // delete mesh;
    return 0;
  }

  int initShadow(){
    glUseProgram(m_program->name);

    GLuint ShadowMapID = glGetUniformLocation(m_program->name, "depth_texture");
    glUniform1i(ShadowMapID, 0);

    glGenFramebuffers(1, &depthFramebuffer);
    glBindFramebuffer(GL_FRAMEBUFFER, depthFramebuffer);

    glGenTextures(1, &depthTexture);
    glBindTexture(GL_TEXTURE_2D, depthTexture);
    glTexImage2D(GL_TEXTURE_2D, 0,GL_DEPTH_COMPONENT16, 512, 512, 0,GL_DEPTH_COMPONENT, GL_FLOAT, 0);
    
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); 
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LEQUAL);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
    
    glFramebufferTexture(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, depthTexture, 0);

    glDrawBuffer(GL_NONE); // No color buffer is drawn to.
 
    if(glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
      return -1;

    return 0;
  }

  int init( ) {
    // compilation simplifiee d'un shader program
    gk::programPath("shaders");
    m_program= gk::createProgram("tp2.glsl");
    if(m_program == gk::GLProgram::null())
      return -1;    
  
    m_shadow= gk::createProgram("shadow.glsl");
    if(m_shadow == gk::GLProgram::null())
      return -1;  
    
    m_sky= gk::createProgram("skymap.glsl");
    if(m_sky == gk::GLProgram::null())
      return -1;  
  
  

    if(initMesh() or initTextures() or initSkyMap() or initShadow())
      return -1;
    
    // for(unsigned t=0;t<stats.size();t++)
    //   std::cout<<"["<<t<<"] = "<<stats[t]<<std::endl;
   

    // nettoyage de l'etat opengl
    glBindVertexArray(0);   // desactive le vertex array
    glBindBuffer(GL_ARRAY_BUFFER, 0);       // desactive le buffer de positions
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);       // desactive le buffer d'indices
    glBindTexture(GL_TEXTURE_2D, 0);

    m_program->uniform("light_position")= light_position;
    m_program->uniform("light_direction")= gk::Vec4(light_direction.x,light_direction.y,light_direction.z,1.0);
    m_program->uniform("light_color")= gk::Vec4(0.5f, 0.5f, 0.5f, 1.0f);  

    m_program->uniform("fog_color")= gk::Vec4(0.5f, 0.5f, 0.5f, 1.0f);  
    m_program->uniform("fog_density")= (float) 0.05;

    // mesure du temps de dessin
    m_time= gk::createTimeQuery();
    mousePress=0;
    update=1;
    // ok, tout s'est bien passe
    return 0;
  }
    


  int quit( )
  {
    return 0;
  }

  // a redefinir pour utiliser les widgets.
  void processWindowResize( SDL_WindowEvent& event )
  {
    m_widgets.reshape(event.data1, event.data2);
  }
    
  // a redefinir pour utiliser les widgets.
  void processMouseButtonEvent( SDL_MouseButtonEvent& event )
  {
    //SDL_PumpEvents();
    if(SDL_GetMouseState(NULL, NULL)&SDL_BUTTON(1))
      mousePress = 1;
    else
      mousePress=0;
    //m_widgets.processMouseButtonEvent(event);
  }
    
  // a redefinir pour utiliser les widgets.
  void processMouseMotionEvent( SDL_MouseMotionEvent& event )
  {
    if(mousePress){
      int xm,ym; 
      SDL_GetRelativeMouseState(&xm, &ym);
      //SDL_GetMouseState(&xm, &ym);
      if(mousePress<2) mousePress++;
      else{
	m_camera.rotate(xm,ym);
	update=1;
	//	SDL_WarpMouseInWindow(m_window,m_width/2,m_height/2);
      }
    }
  }
    
  // a redefinir pour utiliser les widgets.
  void processKeyboardEvent( SDL_KeyboardEvent& event )
  {
    const Uint8 *state = SDL_GetKeyboardState(NULL);
   
    if ( state[SDL_SCANCODE_W] or state[SDL_SCANCODE_UP] ) 
      m_camera.advance();
    if ( state[SDL_SCANCODE_S] or state[SDL_SCANCODE_DOWN] ) 
      m_camera.back();
    if ( state[SDL_SCANCODE_A] or state[SDL_SCANCODE_LEFT] ) 
      m_camera.left();
    if ( state[SDL_SCANCODE_D] or state[SDL_SCANCODE_RIGHT] ) 
      m_camera.right();

    if ( state[SDL_SCANCODE_I] ){ //afficher les infos ...
      shadow_zone.print();
      std::cout<<"Camera position : ";
      m_camera.getPosition().print();
      std::cout<<"Camera direction : ";
      m_camera.getTarget().print();
    }

    if ( state[SDL_SCANCODE_SPACE])
      m_camera.up();

    if ( state[SDL_SCANCODE_LSHIFT])
      m_camera.down();
    
    if ( state[SDL_SCANCODE_R] ) 
      gk::reloadPrograms();  // recharge et recompile les shaders
    if ( state[SDL_SCANCODE_C] ) 
      gk::writeFramebuffer("screenshot.png");	// enregistre l'image opengl
    
    if ( state[SDL_SCANCODE_ESCAPE] ) {
      // fermer l'application si l'utilisateur appuie sur ESCAPE
      printf("Goodbye.\n");
      closeWindow();
    }else
      update=1;
    m_widgets.processKeyboardEvent(event);
  }
    
  int draw( )  {      

    gk::HPoint h; 
    gk::Point p,q;
    gk::Ray ray;
    gk::Vector dir;
    unsigned int i,j, visible;

    gk::Transform viewport   = gk::Viewport(m_width, m_height);

    if(update){

      gk::Point currentBlockTmp = getPointBlock(m_camera.getPosition());
      if(currentBlockTmp.x != currentBlock.x || currentBlockTmp.y != currentBlock.y || currentBlockTmp.z != currentBlock.z){
	currentBlock=currentBlockTmp;
	//ICI : pour mettre à jours le buffer de données
	//loadLocalMash(currentBlock, distanceMaxMap);
      }

      std::array<gk::Vector,8> viewBox = m_camera.getViewBox(); 
      std::array<Plane,6> volume = m_camera.getVolume(); 
      gk::Point pMin,pMax,current, last;
      pMin=gk::Point(99999,99999,99999);
      pMax=gk::Point(-99999,-99999,-99999);
      v_visibleBox.clear();

      //pour chaque point de la bbox compter le nombre de point non visible.
      //si ce nombre est de 8, le block est exclus.
      for(i=0;i<v_pbox.size();i++){
	//pour chaque boite visible calculer la distance entre le block de la caméra et le block
	gk::Point c=getPointBlock(((v_pbox[i][0]+v_pbox[i][1]))*0.5);
	float d = Distance(currentBlock,c);	  
	if(!distanceMaxPrint || distanceMaxPrint >= d){
	  visible = m_camera.visible(v_pbox[i]);	
	  //si le block est visible et
	  //si le nombre de pixel utilisé est superieur à 1
	  if(visible){
	    std::array<gk::Vector,8> pts;
	    double max = 0;
	    for(j=0;j<v_pbox[i].size();j++){
	      //pour chaque sommet de la boite faire mvp, 
	      //faire un tri en X prendre la distance max, pareil en Y.
	      //si la distance max <16 alors inutil d'afficher

	      pts[j] = viewport(m_camera.getMVP()(gk::Point(v_pbox[i][j])));    
	      if(v_pbox[i][j].x<pMin.x)
		pMin.x=v_pbox[i][j].x;
	      if(v_pbox[i][j].y<pMin.y)
		pMin.y=v_pbox[i][j].y;
	      if(v_pbox[i][j].z<pMin.z)
		pMin.z=v_pbox[i][j].z;
	      if(v_pbox[i][j].x>pMax.x)
		pMax.x=v_pbox[i][j].x;
	      if(v_pbox[i][j].y>pMax.y)
		pMax.y=v_pbox[i][j].y;
	      if(v_pbox[i][j].z>pMax.z)
		pMax.z=v_pbox[i][j].z;
	    }
	 
	    std::sort (pts.begin(), pts.end(), compareX);
	    max = fabs(pts[0].x - pts[7].x);
	    std::sort (pts.begin(), pts.end(), compareY);
	    max = std::max(max,fabs(pts[0].y - pts[7].y));
	    if(max>=16)
	      v_visibleBox.push_back(std::make_pair<int,double>(i,d));

	  }
	}
      }
      std::sort (v_visibleBox.begin(), v_visibleBox.end(), compareVisible);
      //      std::cout<<"Draw "<<v_visibleBox.size()<<" blocks."<<std::endl;
      shadow_zone=gk::BBox(pMin,pMax);
      update = 0;    
    
    }
    
    //////////////////////////////////////////////////////////////
    

    gk::Transform depthModel; 
    gk::Transform depthView = gk::LookAt(gk::Point(light_position),
					 shadow_zone.getCenter(),//gk::Point(0.f,0.f,0.f),
					 gk::Vector(0.0f, 1.0f, 0.0f)); 
    //    gk::Transform depthProjection = gk::Orthographic(-50,50,-50,50,-50,70);
    gk::Transform depthProjection= gk::Perspective(50.f, 1.f, 1.f, 1000.f);    
    gk::Transform depthMV= depthView * depthModel ;
    gk::Transform depthMVP= depthProjection * depthMV;
    gk::Transform bias(gk::Matrix4x4(
				     0.5, 0.0, 0.0, 0.0, 
				     0.0, 0.5, 0.0, 0.0,
				     0.0, 0.0, 0.5, 0.0,
				     0.5, 0.5, 0.5, 1.0
				     ));

    gk::Transform depthBiasMVP = bias*depthMVP;


    //gk::Transform depthBiasMVP = depthMVP;

    m_program->uniform("mvpMatrix")= m_camera.getMVP().matrix();
    m_program->uniform("mvMatrix")= m_camera.getMV().matrix();
    m_program->uniform("depthMVP")= depthBiasMVP.matrix();         
    m_program->uniform("camera_position")= m_camera.getPosition();

    m_program->uniform("fogNear")= (float)1.0;
    m_program->uniform("fogFar")= (float)10.0;
    m_program->uniform("fogColor")= gk::Vec3(0.5f, 0.5f, 0.5f);  


    // // dimensions de la fenêtre
  
 
    //~ // mesurer le temps d'execution
    m_time->begin();
    GLint64 start; glGetInteger64v(GL_TIMESTAMP, &start);

    /////////////////////////////////////////////
    glBindFramebuffer(GL_FRAMEBUFFER, depthFramebuffer);
    glViewport(0, 0, m_width, m_height);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    /*
    glUseProgram(m_shadow->name);
     
    
      m_shadow->uniform("depthMVP")= depthMVP.matrix();
      glBindVertexArray(m_vao);
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bindex);
      for(i=0;i<v_visibleBox.size();i++){   
      glDrawElements(GL_TRIANGLES,v_index_size[v_visibleBox[i].first].second, GL_UNSIGNED_INT, reinterpret_cast<void*>( ( v_index_size[v_visibleBox[i].first].first )* sizeof(GLuint) ));
      }
    */
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
     
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK); // Cull back-facing triangles -> draw only front-facing triangles
        
    
 
    //scene
    glViewport(0, 0, m_width, m_height);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


    ///////////////////////////////////////
    
    glDisable(GL_DEPTH_TEST);
    //glDisable(GL_LIGHTING);

    //glFrontFace(GL_CCW);

    glUseProgram(m_sky->name);
    m_sky->uniform("modelview_inv")= m_camera.getMV().inverse().matrix();

    glActiveTexture(GL_TEXTURE5);
    glBindTexture(skyTexture->target, skyTexture->name);

    glBindVertexArray(m_sky_vao);
    glDrawArrays(GL_TRIANGLES, 0, 36);
    //glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_sky_index_buffer->name);
    //glDrawElements(GL_TRIANGLES, m_sky_indices_size, GL_UNSIGNED_INT, 0);
    
    //    glFrontFace(GL_CCW);
    glEnable(GL_DEPTH_TEST);
    //glEnable(GL_LIGHTING);
    
    //////////////////////////////////////////////////////


    glUseProgram(m_program->name);
  
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, depthTexture);
  
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(textures->target, textures->name);


    // glActiveTexture(GL_TEXTURE1);
    //glBindTexture(GL_TEXTURE_CUBE_MAP_ARRAY, textureMapArray);
    //glBindTexture(GL_TEXTURE_CUBE_MAP, textureMapArray);
  
      
    glBindVertexArray(m_vao);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bindex);
    for(i=0;i<v_visibleBox.size();i++){   
      glDrawElements(GL_TRIANGLES,v_index_size[v_visibleBox[i].first].second, GL_UNSIGNED_INT, reinterpret_cast<void*>( ( v_index_size[v_visibleBox[i].first].first )* sizeof(GLuint) ));
    }

    ///////////////////////////////////////
    // nettoyage
    glUseProgram(0);
    glBindVertexArray(0);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
    glBindTexture(GL_TEXTURE_CUBE_MAP_ARRAY, 0);
    //~ // mesurer le temps d'execution
    GLint64 stop; glGetInteger64v(GL_TIMESTAMP, &stop);     // nano secondes
    m_time->end();
    GLuint64 cpu_time= (stop - start) / 1000;               // conversion en micro secondes
    GLuint64 gpu_time= m_time->result64() / 1000;        // conversion en micro secondes
        
    //~ // afficher le temps d'execution
    {

    
      m_widgets.begin();
      m_widgets.beginGroup(nv::GroupFlags_GrowDownFromLeft);
            
      m_widgets.doLabel(nv::Rect(), Format("cpu time % 6ldus", cpu_time));
      m_widgets.doLabel(nv::Rect(), Format("gpu time % 3ldms % 3ldus", gpu_time / 1000, gpu_time % 1000));
      
      frame++;
      time=  SDL_GetTicks();
      if (time - timebase > 1000) {
	fps=frame*1000.0/(time-timebase);

	timebase = time;
	frame = 0;
      }
      m_widgets.doLabel(nv::Rect(), Format("FPS:%4.2f",fps));        
      m_widgets.endGroup();
      m_widgets.end();
    }
    //m_widgets.end();
    // afficher le dessin
    present();
    // continuer
    return 1;
  }
};


int main( int argc, char **argv )
{
  TP app;
  app.run();
    
  return 0;
}

